import * as fs from 'fs';
import * as stream from 'stream';
import readLocalFileHeader from './readLocalFileHeader';

interface IGetFileFromDisksOptions {
  /** Number of the disk where the local file header starts */
  diskStart: number;
  /** Offset into the start of the disk where the local file header starts. */
  offset: number;
  /** Length of the stored file (compressed size + optional 12 byte encryption header), but excluding the length of the local file header. */
  storedSize: number;
}

function getStream(disks: string[], index: number, offset: number, length: number = Infinity) {
  return fs.createReadStream(disks[index], { start: offset, end: offset + length - 1 });
}

/** Takes a list of ReadableStreams (the disks), as well as the offset and length, and returns a stream for just one file. */
export default async function getFileFromDisks(disks: string[], { diskStart, offset, storedSize }: IGetFileFromDisksOptions): Promise<stream.Readable> {
  let curDiskIndex = diskStart;
  let curDisk = getStream(disks, diskStart, offset);
  let localFileHeaderLength = 0;
  let totalRead = 0;

  //Create new stream that concatenates disks until storedSize is reached, then ends the stream.
  const outputStream = new stream.PassThrough();

  const onData = (chunk: Buffer) => {
    totalRead += chunk.length;
    //If we've reached the end, we can stop reading after this chunk
    const readTooManyBytes = totalRead - (localFileHeaderLength + storedSize);
    if (localFileHeaderLength !== 0 && readTooManyBytes >= 0) {
      //We can still write the whole chunk before ending the stream
      if (readTooManyBytes === 0) {
        outputStream.end(chunk);
      } else {
        //We must shorten the chunk, write the shortened chunk and then end the stream
        const shortenedLength = chunk.length - readTooManyBytes;
        const shortenedChunk = new Buffer(shortenedLength);
        shortenedChunk.fill(chunk, 0, shortenedLength);
        outputStream.end(shortenedChunk);
      }
    } else {
      //Nowhere near the end, so just write normally
      outputStream.write(chunk);
    }
  };
  const onError = (error: any) => {
    console.error(error);
  };
  const onEnd = () => {
    curDiskIndex += 1;
    //End if we are at end of file or end of disks
    if (curDiskIndex >= disks.length || (localFileHeaderLength !== 0 && totalRead >= localFileHeaderLength + storedSize)) {
      outputStream.end();
    } else {
      curDisk = getStream(disks, curDiskIndex, 0, (localFileHeaderLength === 0) ? Infinity : localFileHeaderLength + storedSize - totalRead);
      //set up new listeners for data and end
      curDisk.on('data', onData);
      curDisk.on('end', onEnd);
      curDisk.on('error', onError);
    }
  };

  curDisk.on('data', onData);
  curDisk.on('end', onEnd);
  curDisk.on('error', onError);

  //Read local file header
  localFileHeaderLength = await readLocalFileHeader(outputStream);
  //now that local file header has been read, we will restrict length of stream to storedSize

  return outputStream;
}
